# -*- coding: utf-8 -*-

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
from typing import Any, Dict, Optional, Sequence, Union
from google.cloud.aiplatform_v1.types import api_auth
from google.cloud.aiplatform_v1.types import EncryptionSpec
from google.cloud.aiplatform_v1 import (
    RagEmbeddingModelConfig as GapicRagEmbeddingModelConfig,
    GoogleDriveSource,
    ImportRagFilesConfig,
    ImportRagFilesRequest,
    RagEngineConfig as GapicRagEngineConfig,
    RagFileChunkingConfig,
    RagFileParsingConfig,
    RagFileTransformationConfig,
    RagCorpus as GapicRagCorpus,
    RagFile as GapicRagFile,
    RagManagedDbConfig as GapicRagManagedDbConfig,
    SharePointSources as GapicSharePointSources,
    SlackSource as GapicSlackSource,
    JiraSource as GapicJiraSource,
    RagVectorDbConfig as GapicRagVectorDbConfig,
    VertexAiSearchConfig as GapicVertexAiSearchConfig,
)
from google.cloud.aiplatform import initializer
from google.cloud.aiplatform.utils import (
    VertexRagDataAsyncClientWithOverride,
    VertexRagDataClientWithOverride,
    VertexRagClientWithOverride,
)
from vertexai.rag.utils.resources import (
    Basic,
    JiraSource,
    LayoutParserConfig,
    LlmParserConfig,
    Pinecone,
    RagCitedGenerationResponse,
    RagCorpus,
    RagEmbeddingModelConfig,
    RagEngineConfig,
    RagFile,
    RagManagedDb,
    RagManagedDbConfig,
    RagVectorDbConfig,
    Scaled,
    SharePointSources,
    SlackChannelsSource,
    TransformationConfig,
    Unprovisioned,
    VertexAiSearchConfig,
    VertexVectorSearch,
    VertexPredictionEndpoint,
)


_VALID_RESOURCE_NAME_REGEX = "[a-z][a-zA-Z0-9._-]{0,127}"
_VALID_DOCUMENT_AI_PROCESSOR_NAME_REGEX = (
    r"projects/[^/]+/locations/[^/]+/processors/[^/]+(?:/processorVersions/[^/]+)?"
)


def create_rag_data_service_client(
    api_path_override: Optional[str] = None,
):
    return initializer.global_config.create_client(
        client_class=VertexRagDataClientWithOverride,
        api_path_override=api_path_override,
    ).select_version("v1")


def create_rag_data_service_async_client(
    api_path_override: Optional[str] = None,
):
    return initializer.global_config.create_client(
        client_class=VertexRagDataAsyncClientWithOverride,
        api_path_override=api_path_override,
    ).select_version("v1")


def create_rag_service_client(
    api_path_override: Optional[str] = None,
):
    return initializer.global_config.create_client(
        client_class=VertexRagClientWithOverride,
        api_path_override=api_path_override,
    ).select_version("v1")


def convert_gapic_to_rag_embedding_model_config(
    gapic_embedding_model_config: GapicRagEmbeddingModelConfig,
) -> RagEmbeddingModelConfig:
    """Convert GapicRagEmbeddingModelConfig to RagEmbeddingModelConfig."""
    embedding_model_config = RagEmbeddingModelConfig()
    path = gapic_embedding_model_config.vertex_prediction_endpoint.endpoint
    publisher_model = re.match(
        r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/publishers/google/models/(?P<model_id>.+?)$",
        path,
    )
    endpoint = re.match(
        r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/endpoints/(?P<endpoint>.+?)$",
        path,
    )
    if publisher_model:
        embedding_model_config.vertex_prediction_endpoint = VertexPredictionEndpoint(
            publisher_model=path
        )
    if endpoint:
        embedding_model_config.vertex_prediction_endpoint = VertexPredictionEndpoint(
            endpoint=path,
            model=gapic_embedding_model_config.vertex_prediction_endpoint.model,
            model_version_id=gapic_embedding_model_config.vertex_prediction_endpoint.model_version_id,
        )
    return embedding_model_config


def _check_weaviate(gapic_vector_db: GapicRagVectorDbConfig) -> bool:
    try:
        return gapic_vector_db.__contains__("weaviate")
    except AttributeError:
        return gapic_vector_db.weaviate.ByteSize() > 0


def _check_rag_managed_db(gapic_vector_db: GapicRagVectorDbConfig) -> bool:
    try:
        return gapic_vector_db.__contains__("rag_managed_db")
    except AttributeError:
        return gapic_vector_db.rag_managed_db.ByteSize() > 0


def _check_vertex_feature_store(gapic_vector_db: GapicRagVectorDbConfig) -> bool:
    try:
        return gapic_vector_db.__contains__("vertex_feature_store")
    except AttributeError:
        return gapic_vector_db.vertex_feature_store.ByteSize() > 0


def _check_pinecone(gapic_vector_db: GapicRagVectorDbConfig) -> bool:
    try:
        return gapic_vector_db.__contains__("pinecone")
    except AttributeError:
        return gapic_vector_db.pinecone.ByteSize() > 0


def _check_vertex_vector_search(gapic_vector_db: GapicRagVectorDbConfig) -> bool:
    try:
        return gapic_vector_db.__contains__("vertex_vector_search")
    except AttributeError:
        return gapic_vector_db.vertex_vector_search.ByteSize() > 0


def _check_rag_embedding_model_config(
    gapic_vector_db: GapicRagVectorDbConfig,
) -> bool:
    try:
        return gapic_vector_db.__contains__("rag_embedding_model_config")
    except AttributeError:
        return gapic_vector_db.rag_embedding_model_config.ByteSize() > 0


def convert_gapic_to_backend_config(
    gapic_vector_db: GapicRagVectorDbConfig,
) -> RagVectorDbConfig:
    """Convert Gapic RagVectorDbConfig to VertexVectorSearch, Pinecone, or RagManagedDb."""
    vector_config = RagVectorDbConfig()
    if _check_pinecone(gapic_vector_db):
        vector_config.vector_db = Pinecone(
            index_name=gapic_vector_db.pinecone.index_name,
            api_key=gapic_vector_db.api_auth.api_key_config.api_key_secret_version,
        )
    elif _check_vertex_vector_search(gapic_vector_db):
        vector_config.vector_db = VertexVectorSearch(
            index_endpoint=gapic_vector_db.vertex_vector_search.index_endpoint,
            index=gapic_vector_db.vertex_vector_search.index,
        )
    elif _check_rag_managed_db(gapic_vector_db):
        vector_config.vector_db = RagManagedDb()
    if _check_rag_embedding_model_config(gapic_vector_db):
        vector_config.rag_embedding_model_config = (
            convert_gapic_to_rag_embedding_model_config(
                gapic_vector_db.rag_embedding_model_config
            )
        )
    return vector_config


def convert_gapic_to_vertex_ai_search_config(
    gapic_vertex_ai_search_config: VertexAiSearchConfig,
) -> VertexAiSearchConfig:
    """Convert Gapic VertexAiSearchConfig to VertexAiSearchConfig."""
    if gapic_vertex_ai_search_config.serving_config:
        return VertexAiSearchConfig(
            serving_config=gapic_vertex_ai_search_config.serving_config,
        )
    return None


def convert_gapic_to_rag_corpus(gapic_rag_corpus: GapicRagCorpus) -> RagCorpus:
    """Convert GapicRagCorpus to RagCorpus."""
    rag_corpus = RagCorpus(
        name=gapic_rag_corpus.name,
        display_name=gapic_rag_corpus.display_name,
        description=gapic_rag_corpus.description,
        vertex_ai_search_config=convert_gapic_to_vertex_ai_search_config(
            gapic_rag_corpus.vertex_ai_search_config
        ),
        backend_config=convert_gapic_to_backend_config(
            gapic_rag_corpus.vector_db_config
        ),
        encryption_spec=gapic_rag_corpus.encryption_spec,
    )
    return rag_corpus


def convert_gapic_to_rag_corpus_no_embedding_model_config(
    gapic_rag_corpus: GapicRagCorpus,
) -> RagCorpus:
    """Convert GapicRagCorpus without embedding model config (for UpdateRagCorpus) to RagCorpus."""
    rag_vector_db_config_no_embedding_model_config = gapic_rag_corpus.vector_db_config
    rag_vector_db_config_no_embedding_model_config.rag_embedding_model_config = None
    rag_corpus = RagCorpus(
        name=gapic_rag_corpus.name,
        display_name=gapic_rag_corpus.display_name,
        description=gapic_rag_corpus.description,
        vertex_ai_search_config=convert_gapic_to_vertex_ai_search_config(
            gapic_rag_corpus.vertex_ai_search_config
        ),
        backend_config=convert_gapic_to_backend_config(
            rag_vector_db_config_no_embedding_model_config
        ),
        encryption_spec=gapic_rag_corpus.encryption_spec,
    )
    return rag_corpus


def convert_gapic_to_rag_file(gapic_rag_file: GapicRagFile) -> RagFile:
    """Convert GapicRagFile to RagFile."""
    rag_file = RagFile(
        name=gapic_rag_file.name,
        display_name=gapic_rag_file.display_name,
        description=gapic_rag_file.description,
    )
    return rag_file


def convert_json_to_rag_file(upload_rag_file_response: Dict[str, Any]) -> RagFile:
    """Converts a JSON response to a RagFile."""
    rag_file = RagFile(
        name=upload_rag_file_response.get("ragFile").get("name"),
        display_name=upload_rag_file_response.get("ragFile").get("displayName"),
        description=upload_rag_file_response.get("ragFile").get("description"),
    )
    return rag_file


def convert_tuple_to_rag_cited_generation_response(
    cited_text: str, final_bibliography: str
) -> RagCitedGenerationResponse:
    """Converts a tuple to a RagCitedGenerationResponse."""
    rag_cited_generation_response = RagCitedGenerationResponse(
        cited_text=cited_text,
        final_bibliography=final_bibliography,
    )
    return rag_cited_generation_response


def convert_path_to_resource_id(
    path: str,
) -> Union[str, GoogleDriveSource.ResourceId]:
    """Converts a path to a Google Cloud storage uri or GoogleDriveSource.ResourceId."""
    if path.startswith("gs://"):
        # Google Cloud Storage source
        return path
    elif path.startswith("https://drive.google.com/"):
        # Google Drive source
        path_list = path.split("/")
        if "file" in path_list:
            index = path_list.index("file") + 2
            resource_id = path_list[index].split("?")[0]
            resource_type = GoogleDriveSource.ResourceId.ResourceType.RESOURCE_TYPE_FILE
        elif "folders" in path_list:
            index = path_list.index("folders") + 1
            resource_id = path_list[index].split("?")[0]
            resource_type = (
                GoogleDriveSource.ResourceId.ResourceType.RESOURCE_TYPE_FOLDER
            )
        else:
            raise ValueError("path %s is not a valid Google Drive url.", path)

        return GoogleDriveSource.ResourceId(
            resource_id=resource_id,
            resource_type=resource_type,
        )
    else:
        raise ValueError(
            "path must be a Google Cloud Storage uri or a Google Drive url."
        )


def convert_source_for_rag_import(
    source: Union[SlackChannelsSource, JiraSource, SharePointSources],
) -> Union[GapicSlackSource, GapicJiraSource]:
    """Converts a SlackChannelsSource or JiraSource to a GapicSlackSource or GapicJiraSource."""
    if isinstance(source, SlackChannelsSource):
        result_source_channels = []
        for channel in source.channels:
            api_key = channel.api_key
            cid = channel.channel_id
            start_time = channel.start_time
            end_time = channel.end_time
            result_channels = GapicSlackSource.SlackChannels(
                channels=[
                    GapicSlackSource.SlackChannels.SlackChannel(
                        channel_id=cid,
                        start_time=start_time,
                        end_time=end_time,
                    )
                ],
                api_key_config=api_auth.ApiAuth.ApiKeyConfig(
                    api_key_secret_version=api_key
                ),
            )
            result_source_channels.append(result_channels)
        return GapicSlackSource(
            channels=result_source_channels,
        )
    elif isinstance(source, JiraSource):
        result_source_queries = []
        for query in source.queries:
            api_key = query.api_key
            custom_queries = query.custom_queries
            projects = query.jira_projects
            email = query.email
            server_uri = query.server_uri
            result_query = GapicJiraSource.JiraQueries(
                custom_queries=custom_queries,
                projects=projects,
                email=email,
                server_uri=server_uri,
                api_key_config=api_auth.ApiAuth.ApiKeyConfig(
                    api_key_secret_version=api_key
                ),
            )
            result_source_queries.append(result_query)
        return GapicJiraSource(
            jira_queries=result_source_queries,
        )
    elif isinstance(source, SharePointSources):
        result_source_share_point_sources = []
        for share_point_source in source.share_point_sources:
            sharepoint_folder_path = share_point_source.sharepoint_folder_path
            sharepoint_folder_id = share_point_source.sharepoint_folder_id
            drive_name = share_point_source.drive_name
            drive_id = share_point_source.drive_id
            client_id = share_point_source.client_id
            client_secret = share_point_source.client_secret
            tenant_id = share_point_source.tenant_id
            sharepoint_site_name = share_point_source.sharepoint_site_name
            result_share_point_source = GapicSharePointSources.SharePointSource(
                client_id=client_id,
                client_secret=api_auth.ApiAuth.ApiKeyConfig(
                    api_key_secret_version=client_secret
                ),
                tenant_id=tenant_id,
                sharepoint_site_name=sharepoint_site_name,
            )
            if sharepoint_folder_path is not None and sharepoint_folder_id is not None:
                raise ValueError(
                    "sharepoint_folder_path and sharepoint_folder_id cannot both be set."
                )
            elif sharepoint_folder_path is not None:
                result_share_point_source.sharepoint_folder_path = (
                    sharepoint_folder_path
                )
            elif sharepoint_folder_id is not None:
                result_share_point_source.sharepoint_folder_id = sharepoint_folder_id
            if drive_name is not None and drive_id is not None:
                raise ValueError("drive_name and drive_id cannot both be set.")
            elif drive_name is not None:
                result_share_point_source.drive_name = drive_name
            elif drive_id is not None:
                result_share_point_source.drive_id = drive_id
            else:
                raise ValueError("Either drive_name and drive_id must be set.")
            result_source_share_point_sources.append(result_share_point_source)
        return GapicSharePointSources(
            share_point_sources=result_source_share_point_sources,
        )
    else:
        raise TypeError(
            "source must be a SlackChannelsSource or JiraSource or SharePointSources."
        )


def prepare_import_files_request(
    corpus_name: str,
    paths: Optional[Sequence[str]] = None,
    source: Optional[Union[SlackChannelsSource, JiraSource, SharePointSources]] = None,
    transformation_config: Optional[TransformationConfig] = None,
    max_embedding_requests_per_min: int = 1000,
    import_result_sink: Optional[str] = None,
    partial_failures_sink: Optional[str] = None,
    layout_parser: Optional[LayoutParserConfig] = None,
    llm_parser: Optional[LlmParserConfig] = None,
) -> ImportRagFilesRequest:
    if len(corpus_name.split("/")) != 6:
        raise ValueError(
            "corpus_name must be of the format `projects/{project}/locations/{location}/ragCorpora/{rag_corpus}`"
        )

    rag_file_parsing_config = RagFileParsingConfig()
    if layout_parser is not None:
        if (
            re.fullmatch(
                _VALID_DOCUMENT_AI_PROCESSOR_NAME_REGEX,
                layout_parser.processor_name,
            )
            is None
        ):
            raise ValueError(
                "processor_name must be of the format"
                " `projects/{project_id}/locations/{location}/processors/{processor_id}`or"
                " `projects/{project_id}/locations/{location}/processors/{processor_id}/processorVersions/{processor_version_id}`,"
                f" got {layout_parser.processor_name!r}"
            )
        rag_file_parsing_config.layout_parser = RagFileParsingConfig.LayoutParser(
            processor_name=layout_parser.processor_name,
            max_parsing_requests_per_min=layout_parser.max_parsing_requests_per_min,
        )
    if llm_parser is not None:
        rag_file_parsing_config.llm_parser = RagFileParsingConfig.LlmParser(
            model_name=llm_parser.model_name
        )
        if llm_parser.max_parsing_requests_per_min is not None:
            rag_file_parsing_config.llm_parser.max_parsing_requests_per_min = (
                llm_parser.max_parsing_requests_per_min
            )
        if llm_parser.custom_parsing_prompt is not None:
            rag_file_parsing_config.llm_parser.custom_parsing_prompt = (
                llm_parser.custom_parsing_prompt
            )

    chunk_size = 1024
    chunk_overlap = 200
    if transformation_config and transformation_config.chunking_config:
        chunk_size = transformation_config.chunking_config.chunk_size
        chunk_overlap = transformation_config.chunking_config.chunk_overlap

    rag_file_transformation_config = RagFileTransformationConfig(
        rag_file_chunking_config=RagFileChunkingConfig(
            fixed_length_chunking=RagFileChunkingConfig.FixedLengthChunking(
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap,
            ),
        ),
    )

    import_rag_files_config = ImportRagFilesConfig(
        rag_file_transformation_config=rag_file_transformation_config,
        rag_file_parsing_config=rag_file_parsing_config,
        max_embedding_requests_per_min=max_embedding_requests_per_min,
    )

    import_result_sink = import_result_sink or partial_failures_sink

    if import_result_sink is not None:
        if import_result_sink.startswith("gs://"):
            import_rag_files_config.partial_failure_gcs_sink.output_uri_prefix = (
                import_result_sink
            )
        elif import_result_sink.startswith("bq://"):
            import_rag_files_config.partial_failure_bigquery_sink.output_uri = (
                import_result_sink
            )
        else:
            raise ValueError(
                "import_result_sink must be a GCS path or a BigQuery table."
            )

    if source is not None:
        gapic_source = convert_source_for_rag_import(source)
        if isinstance(gapic_source, GapicSlackSource):
            import_rag_files_config.slack_source = gapic_source
        if isinstance(gapic_source, GapicJiraSource):
            import_rag_files_config.jira_source = gapic_source
        if isinstance(gapic_source, GapicSharePointSources):
            import_rag_files_config.share_point_sources = gapic_source
    else:
        uris = []
        resource_ids = []
        for p in paths:
            output = convert_path_to_resource_id(p)
            if isinstance(output, str):
                uris.append(p)
            else:
                resource_ids.append(output)
        if uris:
            import_rag_files_config.gcs_source.uris = uris
        if resource_ids:
            google_drive_source = GoogleDriveSource(
                resource_ids=resource_ids,
            )
            import_rag_files_config.google_drive_source = google_drive_source

    request = ImportRagFilesRequest(
        parent=corpus_name, import_rag_files_config=import_rag_files_config
    )
    return request


def get_corpus_name(
    name: str,
) -> str:
    if name:
        client = create_rag_data_service_client()
        if client.parse_rag_corpus_path(name):
            return name
        elif re.match("^{}$".format(_VALID_RESOURCE_NAME_REGEX), name):
            return client.rag_corpus_path(
                project=initializer.global_config.project,
                location=initializer.global_config.location,
                rag_corpus=name,
            )
        else:
            raise ValueError(
                "name must be of the format `projects/{project}/locations/{location}/ragCorpora/{rag_corpus}` or `{rag_corpus}`"
            )
    return name


def get_file_name(
    name: str,
    corpus_name: str,
) -> str:
    client = create_rag_data_service_client()
    if client.parse_rag_file_path(name):
        return name
    elif re.match("^{}$".format(_VALID_RESOURCE_NAME_REGEX), name):
        if not corpus_name:
            raise ValueError(
                "corpus_name must be provided if name is a `{rag_file}`, not a "
                "full resource name (`projects/{project}/locations/{location}/ragCorpora/{rag_corpus}/ragFiles/{rag_file}`). "
            )
        return client.rag_file_path(
            project=initializer.global_config.project,
            location=initializer.global_config.location,
            rag_corpus=get_corpus_name(corpus_name),
            rag_file=name,
        )
    else:
        raise ValueError(
            "name must be of the format `projects/{project}/locations/{location}/ragCorpora/{rag_corpus}/ragFiles/{rag_file}` or `{rag_file}`"
        )


def set_embedding_model_config(
    embedding_model_config: RagEmbeddingModelConfig,
    rag_corpus: GapicRagCorpus,
) -> None:
    if embedding_model_config.vertex_prediction_endpoint is None:
        return
    if (
        embedding_model_config.vertex_prediction_endpoint.publisher_model
        and embedding_model_config.vertex_prediction_endpoint.endpoint
    ):
        raise ValueError("publisher_model and endpoint cannot be set at the same time.")
    if (
        not embedding_model_config.vertex_prediction_endpoint.publisher_model
        and not embedding_model_config.vertex_prediction_endpoint.endpoint
    ):
        raise ValueError("At least one of publisher_model and endpoint must be set.")
    parent = initializer.global_config.common_location_path(project=None, location=None)

    if embedding_model_config.vertex_prediction_endpoint.publisher_model:
        publisher_model = (
            embedding_model_config.vertex_prediction_endpoint.publisher_model
        )
        full_resource_name = re.match(
            r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/publishers/google/models/(?P<model_id>.+?)$",
            publisher_model,
        )
        resource_name = re.match(
            r"^publishers/google/models/(?P<model_id>.+?)$",
            publisher_model,
        )
        if full_resource_name:
            rag_corpus.vector_db_config.rag_embedding_model_config.vertex_prediction_endpoint.endpoint = (
                publisher_model
            )
        elif resource_name:
            rag_corpus.vector_db_config.rag_embedding_model_config.vertex_prediction_endpoint.endpoint = (
                parent + "/" + publisher_model
            )
        else:
            raise ValueError(
                "publisher_model must be of the format `projects/{project}/locations/{location}/publishers/google/models/{model_id}` or `publishers/google/models/{model_id}`"
            )

    if embedding_model_config.vertex_prediction_endpoint.endpoint:
        endpoint = embedding_model_config.vertex_prediction_endpoint.endpoint
        full_resource_name = re.match(
            r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/endpoints/(?P<endpoint>.+?)$",
            endpoint,
        )
        resource_name = re.match(
            r"^endpoints/(?P<endpoint>.+?)$",
            endpoint,
        )
        if full_resource_name:
            rag_corpus.vector_db_config.rag_embedding_model_config.vertex_prediction_endpoint.endpoint = (
                endpoint
            )
        elif resource_name:
            rag_corpus.vector_db_config.rag_embedding_model_config.vertex_prediction_endpoint.endpoint = (
                parent + "/" + endpoint
            )
        else:
            raise ValueError(
                "endpoint must be of the format `projects/{project}/locations/{location}/endpoints/{endpoint}` or `endpoints/{endpoint}`"
            )


def set_backend_config(
    backend_config: Optional[
        Union[
            RagVectorDbConfig,
            None,
        ]
    ],
    rag_corpus: GapicRagCorpus,
) -> None:
    """Sets the vector db configuration for the rag corpus."""
    if backend_config is None:
        return

    if backend_config.vector_db is not None:
        vector_config = backend_config.vector_db
        if vector_config is None or isinstance(vector_config, RagManagedDb):
            rag_corpus.vector_db_config.rag_managed_db.CopyFrom(
                GapicRagVectorDbConfig.RagManagedDb()
            )
        elif isinstance(vector_config, VertexVectorSearch):
            index_endpoint = vector_config.index_endpoint
            index = vector_config.index

            rag_corpus.vector_db_config.vertex_vector_search.index_endpoint = (
                index_endpoint
            )
            rag_corpus.vector_db_config.vertex_vector_search.index = index
        elif isinstance(vector_config, Pinecone):
            index_name = vector_config.index_name
            api_key = vector_config.api_key

            rag_corpus.vector_db_config.pinecone.index_name = index_name
            rag_corpus.vector_db_config.api_auth.api_key_config.api_key_secret_version = (
                api_key
            )
        else:
            raise TypeError(
                "backend_config must be a VertexFeatureStore,"
                "RagManagedDb, or Pinecone."
            )
    if backend_config.rag_embedding_model_config:
        set_embedding_model_config(
            backend_config.rag_embedding_model_config, rag_corpus
        )


def set_encryption_spec(
    encryption_spec: EncryptionSpec,
    rag_corpus: GapicRagCorpus,
) -> None:
    """Sets the encryption spec for the rag corpus."""
    # Raises value error if encryption_spec.kms_key_name is None or empty,
    if encryption_spec.kms_key_name is None or not encryption_spec.kms_key_name:
        raise ValueError("kms_key_name must be set if encryption_spec is set.")

    # Raises value error if encryption_spec.kms_key_name is not a valid KMS key name.
    if not re.match(
        r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/keyRings/(?P<key_ring>.+?)/cryptoKeys/(?P<crypto_key>.+?)$",
        encryption_spec.kms_key_name,
    ):
        raise ValueError(
            "kms_key_name must be of the format "
            "`projects/{project}/locations/{location}/keyRings/{key_ring}/cryptoKeys/{crypto_key}`"
        )

    rag_corpus.encryption_spec = encryption_spec


def set_vertex_ai_search_config(
    vertex_ai_search_config: VertexAiSearchConfig,
    rag_corpus: GapicRagCorpus,
) -> None:
    if not vertex_ai_search_config.serving_config:
        raise ValueError("serving_config must be set.")
    engine_resource_name = re.match(
        r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/collections/(?P<collection>.+?)/engines/(?P<engine>.+?)/servingConfigs/(?P<serving_config>.+?)$",
        vertex_ai_search_config.serving_config,
    )
    data_store_resource_name = re.match(
        r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/collections/(?P<collection>.+?)/dataStores/(?P<data_store>.+?)/servingConfigs/(?P<serving_config>.+?)$",
        vertex_ai_search_config.serving_config,
    )
    if engine_resource_name or data_store_resource_name:
        rag_corpus.vertex_ai_search_config = GapicVertexAiSearchConfig(
            serving_config=vertex_ai_search_config.serving_config,
        )
    else:
        raise ValueError(
            "serving_config must be of the format `projects/{project}/locations/{location}/collections/{collection}/engines/{engine}/servingConfigs/{serving_config}` or `projects/{project}/locations/{location}/collections/{collection}/dataStores/{data_store}/servingConfigs/{serving_config}`"
        )


def convert_gapic_to_rag_engine_config(
    response: GapicRagEngineConfig,
) -> RagEngineConfig:
    """Converts a GapicRagEngineConfig to a RagEngineConfig."""
    rag_managed_db_config = RagManagedDbConfig()
    # If future fields are added with similar names, beware that __contains__
    # may match them.
    if response.rag_managed_db_config.__contains__("basic"):
        rag_managed_db_config.tier = Basic()
    elif response.rag_managed_db_config.__contains__("unprovisioned"):
        rag_managed_db_config.tier = Unprovisioned()
    elif response.rag_managed_db_config.__contains__("scaled"):
        rag_managed_db_config.tier = Scaled()
    else:
        raise ValueError("At least one of rag_managed_db_config must be set.")
    return RagEngineConfig(
        name=response.name,
        rag_managed_db_config=rag_managed_db_config,
    )


def convert_rag_engine_config_to_gapic(
    rag_engine_config: RagEngineConfig,
) -> GapicRagEngineConfig:
    """Converts a RagEngineConfig to a GapicRagEngineConfig."""
    rag_managed_db_config = GapicRagManagedDbConfig()
    if (
        rag_engine_config.rag_managed_db_config is None
        or rag_engine_config.rag_managed_db_config.tier is None
    ):
        rag_managed_db_config = GapicRagManagedDbConfig(
            basic=GapicRagManagedDbConfig.Basic()
        )
    else:
        if isinstance(rag_engine_config.rag_managed_db_config.tier, Basic):
            rag_managed_db_config.basic = GapicRagManagedDbConfig.Basic()
        elif isinstance(rag_engine_config.rag_managed_db_config.tier, Unprovisioned):
            rag_managed_db_config.unprovisioned = (
                GapicRagManagedDbConfig.Unprovisioned()
            )
        elif isinstance(rag_engine_config.rag_managed_db_config.tier, Scaled):
            rag_managed_db_config.scaled = GapicRagManagedDbConfig.Scaled()
    return GapicRagEngineConfig(
        name=rag_engine_config.name,
        rag_managed_db_config=rag_managed_db_config,
    )



# -*- coding: utf-8 -*-

# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions for metrics."""

import datetime
import io
from typing import Any, Dict, List, Optional, Union, Callable

from google.cloud.aiplatform import initializer
from google.cloud.aiplatform_v1beta1.types import (
    evaluation_service as gapic_eval_service_types,
)
from vertexai.preview.evaluation import metrics
from vertexai.preview.evaluation import prompt_template
from vertexai.preview.evaluation import utils
from vertexai.preview.evaluation.metrics import _schema
from vertexai.preview.evaluation.metrics import (
    custom_output_config,
)
from vertexai import generative_models
from vertexai.generative_models import _generative_models
import jsonschema
import ruamel.yaml
from ruamel.yaml import scalarstring


AutoraterConfig = gapic_eval_service_types.AutoraterConfig
GenerativeModel = generative_models.GenerativeModel
CustomOutputConfig = custom_output_config.CustomOutputConfig
PairwiseMetric = metrics.PairwiseMetric
PointwiseMetric = metrics.PointwiseMetric
RubricBasedMetric = metrics.RubricBasedMetric
RubricGenerationConfig = metrics.RubricGenerationConfig
PromptTemplate = prompt_template.PromptTemplate

# Initialize schema validator.
_schema = ruamel.yaml.YAML(typ="safe").load(_schema.AUTORATER_METRIC_SCHEMA)
_schema_validator = jsonschema.Draft202012Validator(schema=_schema)


def dump(
    metric: Union[PointwiseMetric, PairwiseMetric, RubricBasedMetric],
    file_path: str,
    version: Optional[str] = None,
):
    """Dumps a metric object to a YAML file.

    Args:
      metric: The metric to be dumped to a file.
      file_path: The path to the file. Local and GCS files are supported.
      version: Optional. The version of the metric. Defaults to the timestamp
        when the metric file is created.
    """
    yaml_data = dumps(metric, version)
    if file_path.startswith(utils._GCS_PREFIX):
        utils._upload_string_to_gcs(file_path, yaml_data)
    else:
        with open(file_path, "w") as f:
            f.write(yaml_data)


def dumps(
    metric: Union[PointwiseMetric, PairwiseMetric, RubricBasedMetric],
    version: Optional[str] = None,
) -> str:
    """Dumps a metric object to YAML data.

    Args:
      metric: The metric to be dumped to YAML data.
      version: Optional. The version of the metric. Defaults to the timestamp
        when the metric file is created.

    Returns:
      The YAML data of the metric.
    """
    steps = []
    metric_name = None
    if isinstance(metric, PointwiseMetric) or isinstance(metric, PairwiseMetric):
        metric_name = metric.metric_name
        steps.append(_dump_metric(metric))
    elif isinstance(metric, RubricBasedMetric):
        metric_name = metric.critique_metric.metric_name
        steps.append(_dump_rubric(metric.generation_config))
        steps.append(_dump_metric(metric.critique_metric))
    metadata = {
        "name": metric_name,
        "version": (
            datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            if version is None
            else version
        ),
        "required_inputs": _parse_required_inputs(metric),
    }
    metric_config = {
        "metadata": metadata,
        "steps": steps,
    }
    yaml = ruamel.yaml.YAML()
    yaml.indent(sequence=4, offset=2)
    with io.StringIO() as s:
        yaml.dump(metric_config, s)
        return s.getvalue()


def _dump_metric(metric: Union[PointwiseMetric, PairwiseMetric]) -> Dict[str, Any]:
    """Dumps a metric object to autorater metric schema."""
    output_type = None
    if metric.custom_output_config and metric.custom_output_config.return_raw_output:
        output_type = "raw"
    step = {
        "type": (
            "pairwise_metric"
            if isinstance(metric, PairwiseMetric)
            else "pointwise_metric"
        ),
        "prompt": {
            "template": scalarstring.preserve_literal(metric.metric_prompt_template),
        },
    }
    if metric.system_instruction:
        step["prompt"]["system_instruction"] = metric.system_instruction
    if output_type:
        step["output"] = {
            "type": output_type,
        }
    if metric.autorater_config:
        step["model"] = {
            "model_name_or_endpoint": (metric.autorater_config.autorater_model),
        }
        options = {}
        if metric.autorater_config.flip_enabled:
            options["flip_enabled"] = metric.autorater_config.flip_enabled
        if metric.autorater_config.sampling_count:
            options["sample_count"] = metric.autorater_config.sampling_count
        if options:
            step["options"] = options
    return step


def _dump_rubric(generation_config: RubricGenerationConfig) -> Dict[str, Any]:
    """Dumps a rubric generation config to autorater metric schema."""
    # TODO: b/396217889 - add support for custom output.
    step = {
        "type": "rubric",
        "prompt": {
            "template": scalarstring.preserve_literal(
                generation_config.prompt_template
            ),
        },
    }
    if generation_config.model and isinstance(generation_config.model, GenerativeModel):
        step["model"] = {
            "model_name_or_endpoint": generation_config.model._model_name,
        }
    return step


def _parse_required_inputs(
    metric: Union[PointwiseMetric, PairwiseMetric, RubricBasedMetric],
) -> List[str]:
    """Parses required inputs from a metric object."""
    if isinstance(metric, PointwiseMetric) or isinstance(metric, PairwiseMetric):
        return list(PromptTemplate(metric.metric_prompt_template).variables)
    elif isinstance(metric, RubricBasedMetric):
        met = PromptTemplate(metric.critique_metric.metric_prompt_template).variables
        gen = PromptTemplate(metric.generation_config.prompt_template).variables
        return list(met.union(gen))
    else:
        raise ValueError(f"Unsupported metric type: {type(metric)}")


def load(
    file_path: str,
    baseline_model: Optional[Union[GenerativeModel, Callable[[str], str]]] = None,
) -> Union[PointwiseMetric, PairwiseMetric, RubricBasedMetric]:
    """Loads a metric object from a YAML file.

    Args:
      file_path: Path to the file containing the autorater metric configuration.
        Local and GCS files are supported.
      baseline_model: Optional. The baseline model to use for pairwise metrics.

    Returns:
      The metric object loaded from the file.
    """
    if file_path.startswith(utils._GCS_PREFIX):
        file_contents = utils._read_gcs_file_contents(file_path)
        return loads(file_contents, baseline_model)
    with open(file_path, "r") as f:
        return loads(f.read(), baseline_model)


def loads(
    yaml_data: str,
    baseline_model: Optional[Union[GenerativeModel, Callable[[str], str]]] = None,
) -> Union[PointwiseMetric, PairwiseMetric, RubricBasedMetric]:
    """Loads a metric object from YAML data.

    Args:
      yaml_data: YAML data containing the autorater metric configuration.
      baseline_model: Optional. The baseline model to use for pairwise metrics.

    Returns:
      The metric object loaded from the YAML data.
    """
    yaml = ruamel.yaml.YAML(typ="safe")
    yaml_obj = yaml.load(yaml_data)
    try:
        _schema_validator.validate(yaml_obj)
    except jsonschema.exceptions.ValidationError as e:
        raise ValueError(
            f"Invalid autorater metric config: {e.message} for {e.path.pop()}"
        ) from e
    metadata = yaml_obj["metadata"]
    steps = yaml_obj["steps"]
    required_inputs = set(metadata["required_inputs"])
    metric = None
    rubric = None
    for step in steps:
        _validate_template(step["prompt"]["template"], required_inputs)
        model_name = None
        flip = None
        sampling = None
        if "model" in step:
            model_name = _parse_model_name(step["model"]["model_name_or_endpoint"])
        if "options" in step:
            flip = step["options"].get("flip_enabled", False)
            sampling = step["options"].get("sample_count", 1)
        autorater = None
        if model_name:
            autorater = AutoraterConfig(
                autorater_model=model_name,
                flip_enabled=flip,
                sampling_count=sampling,
            )
        system_instruction = step["prompt"].get("system_instruction")
        custom_output = None
        if "output" in step and step["output"]["type"] == "raw":
            custom_output = CustomOutputConfig(return_raw_output=True)
        if step["type"] == "pointwise_metric":
            if metric is not None:
                raise ValueError("Only one metric step is supported.")
            if baseline_model:
                raise ValueError("Baseline model provided for pointwise metric.")
            metric = PointwiseMetric(
                metric=metadata["name"],
                metric_prompt_template=step["prompt"]["template"],
                system_instruction=system_instruction,
                autorater_config=autorater,
                custom_output_config=custom_output,
            )
        elif step["type"] == "pairwise_metric":
            if metric is not None:
                raise ValueError("Only one metric step is supported.")
            metric = PairwiseMetric(
                metric=metadata["name"],
                metric_prompt_template=step["prompt"]["template"],
                system_instruction=system_instruction,
                baseline_model=baseline_model,
                autorater_config=autorater,
                custom_output_config=custom_output,
            )
        elif step["type"] == "rubric":
            if rubric is not None:
                raise ValueError("Only one rubric step is supported.")
            model = None
            if model_name:
                model = generative_models.GenerativeModel(model_name=model_name)
            rubric = RubricGenerationConfig(
                prompt_template=step["prompt"]["template"],
                model=model,
            )
    if metric is None:
        raise ValueError("A metric step must be provided.")
    if rubric is not None:
        return RubricBasedMetric(
            generation_config=rubric,
            critique_metric=metric,
        )
    return metric


def _parse_model_name(model_name_or_endpoint: str) -> str:
    """Parses model name or endpoint.

    Args:
      model_name_or_endpoint: Model Garden model  name or tuned model endpoint
        resource name can be provided.

    Returns:
      The model resource name.
    """
    project = initializer.global_config.project
    location = initializer.global_config.location
    model_name = _generative_models._reconcile_model_name(
        model_name_or_endpoint, project, location
    )
    return _generative_models._get_resource_name_from_model_name(
        model_name, project, location
    )


def _validate_template(template: str, required_inputs: List[str]) -> None:
    """Validates the template contains only required inputs."""
    placeholders = PromptTemplate(template).variables
    if not placeholders.issubset(required_inputs):
        raise ValueError(
            "Template contains placeholders that are not in required inputs:"
            f" {placeholders - required_inputs}"
        )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass
from typing import Literal

from opentelemetry.resourcedetector.gcp_resource_detector import (
    _gce,
    _metadata,
)

KUBERNETES_SERVICE_HOST_ENV = "KUBERNETES_SERVICE_HOST"


def on_gke() -> bool:
    return os.environ.get(KUBERNETES_SERVICE_HOST_ENV) is not None


def host_id() -> str:
    return _gce.host_id()


def cluster_name() -> str:
    return _metadata.get_metadata()["instance"]["attributes"]["cluster-name"]


@dataclass
class ZoneOrRegion:
    type: Literal["zone", "region"]
    value: str


def availability_zone_or_region() -> ZoneOrRegion:
    cluster_location = _metadata.get_metadata()["instance"]["attributes"][
        "cluster-location"
    ]
    hyphen_count = cluster_location.count("-")
    if hyphen_count == 1:
        return ZoneOrRegion(type="region", value=cluster_location)
    if hyphen_count == 2:
        return ZoneOrRegion(type="zone", value=cluster_location)
    raise Exception(
        f"unrecognized format for cluster location: {cluster_location}"
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Shared utilities for working with function schemas."""

import inspect
import typing
from typing import Any, Callable, Dict
import warnings

from google.cloud.aiplatform_v1beta1 import types as aiplatform_types

Struct = Dict[str, Any]


def _generate_json_schema_from_function_using_pydantic(
    func: Callable,
) -> Struct:
    """Generates JSON Schema for a callable object.

    The `func` function needs to follow specific rules.
    All parameters must be names explicitly (`*args` and `**kwargs` are not supported).

    Args:
        func: Function for which to generate schema

    Returns:
        The JSON Schema for the function as a dict.
    """
    # FIX(b/331534434): Workaround for a breaking change.
    try:
        from pydantic import v1 as pydantic
        from pydantic.v1 import fields as pydantic_fields
    except ImportError:
        import pydantic
        from pydantic import fields as pydantic_fields

    try:
        import docstring_parser  # pylint: disable=g-import-not-at-top
    except ImportError:
        warnings.warn("Unable to import docstring_parser")
        docstring_parser = None

    function_description = func.__doc__

    # Parse parameter descriptions from the docstring.
    # Also parse the function descripton in a better way.
    parameter_descriptions = {}
    if docstring_parser:
        parsed_docstring = docstring_parser.parse(func.__doc__)
        function_description = (
            parsed_docstring.long_description or parsed_docstring.short_description
        )
        for meta in parsed_docstring.meta:
            if isinstance(meta, docstring_parser.DocstringParam):
                parameter_descriptions[meta.arg_name] = meta.description

    defaults = dict(inspect.signature(func).parameters)
    fields_dict = {
        name: (
            # 1. We infer the argument type here: use Any rather than None so
            # it will not try to auto-infer the type based on the default value.
            (param.annotation if param.annotation != inspect.Parameter.empty else Any),
            pydantic.Field(
                # 2. We do not support default values for now.
                default=(
                    param.default
                    if param.default != inspect.Parameter.empty
                    # ! Need to use Undefined instead of None
                    else pydantic_fields.Undefined
                ),
                # 3. We support user-provided descriptions.
                description=parameter_descriptions.get(name, None),
            ),
        )
        for name, param in defaults.items()
        # We do not support *args or **kwargs
        if param.kind
        in (
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
            inspect.Parameter.POSITIONAL_ONLY,
        )
    }
    function_schema = pydantic.create_model(func.__name__, **fields_dict).schema()

    function_schema["title"] = func.__name__
    function_schema["description"] = function_description
    # Postprocessing
    for name, property_schema in function_schema.get("properties", {}).items():
        annotation = defaults[name].annotation
        # 5. Nullable fields:
        #     * https://github.com/pydantic/pydantic/issues/1270
        #     * https://stackoverflow.com/a/58841311
        #     * https://github.com/pydantic/pydantic/discussions/4872
        if typing.get_origin(annotation) is typing.Union and type(
            None
        ) in typing.get_args(annotation):
            # for "typing.Optional" arguments, function_arg might be a
            # dictionary like
            #
            #   {'anyOf': [{'type': 'integer'}, {'type': 'null'}]
            for schema in property_schema.pop("anyOf", []):
                schema_type = schema.get("type")
                if schema_type and schema_type != "null":
                    property_schema["type"] = schema_type
                    break
            property_schema["nullable"] = True
    # 6. Annotate required fields.
    function_schema["required"] = [
        k
        for k in defaults
        if (
            defaults[k].default == inspect.Parameter.empty
            and defaults[k].kind
            in (
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.KEYWORD_ONLY,
                inspect.Parameter.POSITIONAL_ONLY,
            )
        )
    ]
    return function_schema


def adapt_json_schema_to_google_tool_schema(schema: Struct) -> Struct:
    """Adapts JSON schema to Google tool schema."""
    fixed_schema = dict(schema)
    # `$schema` is one of the basic/most common fields of the real JSON Schema.
    # But Google's Schema proto does not support it.
    # Common attributes that we remove:
    # $schema, additionalProperties
    for key in list(fixed_schema):
        # Warning: The proto-plus library removes attributes from the class,
        # so `hasattr` does not work.
        if (
            key not in aiplatform_types.Schema.meta.fields
            and key + "_" not in aiplatform_types.Schema.meta.fields
        ):
            fixed_schema.pop(key, None)
    property_schemas = fixed_schema.get("properties")
    if property_schemas:
        for k, v in property_schemas.items():
            property_schemas[k] = adapt_json_schema_to_google_tool_schema(v)
    return fixed_schema


generate_json_schema_from_function = _generate_json_schema_from_function_using_pydantic



# Copyright 2016 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common logging helpers."""

import logging

from datetime import datetime
from datetime import timedelta
from datetime import timezone

import requests

from google.cloud.logging_v2.entries import LogEntry
from google.cloud.logging_v2.entries import ProtobufEntry
from google.cloud.logging_v2.entries import StructEntry
from google.cloud.logging_v2.entries import TextEntry

try:
    from google.cloud.logging_v2.types import LogSeverity
except ImportError:  # pragma: NO COVER

    class LogSeverity(object):
        """Map severities for non-GAPIC usage."""

        DEFAULT = 0
        DEBUG = 100
        INFO = 200
        NOTICE = 300
        WARNING = 400
        ERROR = 500
        CRITICAL = 600
        ALERT = 700
        EMERGENCY = 800


_NORMALIZED_SEVERITIES = {
    logging.CRITICAL: LogSeverity.CRITICAL,
    logging.ERROR: LogSeverity.ERROR,
    logging.WARNING: LogSeverity.WARNING,
    logging.INFO: LogSeverity.INFO,
    logging.DEBUG: LogSeverity.DEBUG,
    logging.NOTSET: LogSeverity.DEFAULT,
}

_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"
"""Time format for timestamps used in API"""

METADATA_URL = "http://metadata.google.internal./computeMetadata/v1/"
METADATA_HEADERS = {"Metadata-Flavor": "Google"}


def entry_from_resource(resource, client, loggers):
    """Detect correct entry type from resource and instantiate.

    Args:
        resource (dict): One entry resource from API response.
        client (~logging_v2.client.Client):
            Client that owns the log entry.
        loggers (dict):
            A mapping of logger fullnames -> loggers.  If the logger
            that owns the entry is not in ``loggers``, the entry
            will have a newly-created logger.

    Returns:
        google.cloud.logging_v2.entries._BaseEntry:
            The entry instance, constructed via the resource
    """
    if "textPayload" in resource:
        return TextEntry.from_api_repr(resource, client, loggers=loggers)

    if "jsonPayload" in resource:
        return StructEntry.from_api_repr(resource, client, loggers=loggers)

    if "protoPayload" in resource:
        return ProtobufEntry.from_api_repr(resource, client, loggers=loggers)

    return LogEntry.from_api_repr(resource, client, loggers=loggers)


def retrieve_metadata_server(metadata_key, timeout=5):
    """Retrieve the metadata key in the metadata server.

    See: https://cloud.google.com/compute/docs/storing-retrieving-metadata

    Args:
        metadata_key (str):
            Key of the metadata which will form the url. You can
            also supply query parameters after the metadata key.
            e.g. "tags?alt=json"
        timeout (number): number of seconds to wait for the HTTP request

    Returns:
        str: The value of the metadata key returned by the metadata server.
    """
    url = METADATA_URL + metadata_key

    try:
        response = requests.get(url, headers=METADATA_HEADERS, timeout=timeout)

        if response.status_code == requests.codes.ok:
            return response.text

    except requests.exceptions.RequestException:
        # Ignore the exception, connection failed means the attribute does not
        # exist in the metadata server.
        pass

    return None


def _normalize_severity(stdlib_level):
    """Normalize a Python stdlib severity to LogSeverity enum.

    Args:
        stdlib_level (int): 'levelno' from a :class:`logging.LogRecord`

    Returns:
        int: Corresponding Stackdriver severity.
    """
    return _NORMALIZED_SEVERITIES.get(stdlib_level, stdlib_level)


def _add_defaults_to_filter(filter_):
    """Modify the input filter expression to add sensible defaults.

    Args:
        filter_ (str): The original filter expression

    Returns:
        str: sensible default filter string
    """

    # By default, requests should only return logs in the last 24 hours
    yesterday = datetime.now(timezone.utc) - timedelta(days=1)
    time_filter = f'timestamp>="{yesterday.strftime(_TIME_FORMAT)}"'
    if filter_ is None:
        filter_ = time_filter
    elif "timestamp" not in filter_.lower():
        filter_ = f"{filter_} AND {time_filter}"
    return filter_



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from dataclasses import dataclass

from opentelemetry.resourcedetector.gcp_resource_detector import _metadata

# Format described in
# https://cloud.google.com/compute/docs/metadata/default-metadata-values#vm_instance_metadata
_ZONE_REGION_RE = re.compile(
    r"projects\/\d+\/zones\/(?P<zone>(?P<region>\w+-\w+)-\w+)"
)

_logger = logging.getLogger(__name__)


def on_gce() -> bool:
    try:
        _metadata.get_metadata()["instance"]["machineType"]
    except (_metadata.MetadataAccessException, KeyError):
        _logger.debug(
            "Could not fetch metadata attribute instance/machineType, "
            "assuming not on GCE.",
            exc_info=True,
        )
        return False
    return True


def host_type() -> str:
    return _metadata.get_metadata()["instance"]["machineType"]


def host_id() -> str:
    return str(_metadata.get_metadata()["instance"]["id"])


def host_name() -> str:
    return _metadata.get_metadata()["instance"]["name"]


@dataclass
class ZoneAndRegion:
    zone: str
    region: str


def availability_zone_and_region() -> ZoneAndRegion:
    full_zone = _metadata.get_metadata()["instance"]["zone"]
    match = _ZONE_REGION_RE.search(full_zone)
    if not match:
        raise Exception(
            "zone was not in the expected format: "
            f"projects/PROJECT_NUM/zones/COUNTRY-REGION-ZONE. Got {full_zone}"
        )

    return ZoneAndRegion(
        zone=match.group("zone"), region=match.group("region")
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Live AgentEngine API client."""

import contextlib
import json
from typing import Any, AsyncIterator, Dict, Optional
import google.auth

from google.genai import _api_module
from .types import QueryAgentEngineConfig, QueryAgentEngineConfigOrDict


try:
    from websockets.asyncio.client import ClientConnection
    from websockets.asyncio.client import connect as ws_connect
except ModuleNotFoundError:
    # This try/except is for TAP, mypy complains about it which is why we have the type: ignore
    from websockets.client import ClientConnection  # type: ignore
    from websockets.client import connect as ws_connect  # type: ignore


class AsyncLiveAgentEngineSession:
    """AsyncLiveAgentEngineSession."""

    def __init__(self, websocket: ClientConnection):
        self._ws = websocket

    async def send(self, query_input: Dict[str, Any]) -> None:
        """Send a query input to the Agent.

        Args:
          query_input: A JSON serializable Python Dict to be send to the Agent.
        """

        try:
            json_request = json.dumps({"bidi_stream_input": query_input})
        except json.JSONEncoderError as exc:
            raise ValueError(
                "Failed to encode query input to JSON in live_agent_engines: "
                f"{str(query_input)}"
            ) from exc
        await self._ws.send(json_request)

    async def receive(self) -> Dict[str, Any]:
        """Receive one response from the Agent.

        Returns:
          A response from the Agent.

        Raises:
          websockets.exceptions.ConnectionClosed: If the connection is closed.
        """

        response = await self._ws.recv()
        try:
            return json.loads(response)
        except json.decoder.JSONDecodeError as exc:
            raise ValueError(
                "Failed to parse response to JSON in live_agent_engines: "
                f"{str(response)}"
            ) from exc

    async def close(self) -> None:
        """Close the connection."""
        await self._ws.close()


class AsyncLiveAgentEngines(_api_module.BaseModule):
    """AsyncLiveAgentEngines.

    Example usage:

    .. code-block:: python

      from pathlib import Path

      from google import genai
      from google.genai import types

      class MyAgentEngine(client):
        def bidi_stream_query(self, input_queue: asyncio.Queue):
          while True:
            input = await input_queue.get()
            yield {"output": f"Agent received {input}!"}

      client = vertexai.Client(project="my-project", location="us-central1")
      agent_engine = client.agent_engines.create(agent)

      async with client.aio.live.agent_engines.connect(
          agent_engine=agent_engine.api_resource.name,
          setup={"class_method": "bidi_stream_query"},
      ) as session:
        await session.send(input={"input": "Hello world"})

        response = await session.receive()
        # {"output": "Agent received Hello world!"}
        ...
    """

    @contextlib.asynccontextmanager
    async def connect(
        self,
        *,
        agent_engine: str,
        config: Optional[QueryAgentEngineConfigOrDict] = None,
    ) -> AsyncIterator[AsyncLiveAgentEngineSession]:
        """Connect to the agent deployed to Agent Engine in a live (bidirectional streaming) session.

        Args:
          agent_engine: The resource name of the Agent Engine to use for the
            live session.
          config: The optional configuration for starting the live Agent Engine
            session. Custom class_method and an optional initial input could be
            provided. If no class_method is provided, the default class_method
            "bidi_stream_query" will be used by the Agent Engine.

        Yields:
          An AsyncLiveAgentEngineSession object.
        """
        if isinstance(config, dict):
            config = QueryAgentEngineConfig(**config)

        agent_engine_resource_name = agent_engine
        if not agent_engine_resource_name.startswith("projects/"):
            agent_engine_resource_name = f"projects/{self._api_client.project}/locations/{self._api_client.location}/reasoningEngines/{agent_engine}"
        request_dict = {"setup": {"name": agent_engine_resource_name}}
        if config.class_method:
            request_dict["setup"]["class_method"] = config.class_method
        if config.input:
            request_dict["setup"]["input"] = config.input

        request = json.dumps(request_dict)

        if not self._api_client._credentials:
            # Get bearer token through Application Default Credentials.
            creds, _ = google.auth.default(  # type: ignore
                scopes=["https://www.googleapis.com/auth/cloud-platform"]
            )
        else:
            creds = self._api_client._credentials
        # creds.valid is False, and creds.token is None
        # Need to refresh credentials to populate those
        if not (creds.token and creds.valid):
            auth_req = google.auth.transport.requests.Request()  # type: ignore
            creds.refresh(auth_req)
        bearer_token = creds.token

        original_headers = self._api_client._http_options.headers
        headers = original_headers.copy() if original_headers is not None else {}
        headers["Authorization"] = f"Bearer {bearer_token}"

        base_url = self._api_client._websocket_base_url()
        if isinstance(base_url, bytes):
            base_url = base_url.decode("utf-8")
        uri = (
            f"{base_url}/ws/google.cloud.aiplatform."
            f"{self._api_client._http_options.api_version}"
            ".ReasoningEngineExecutionService/BidiQueryReasoningEngine"
        )

        async with ws_connect(
            uri, additional_headers=headers, **self._api_client._websocket_ssl_ctx
        ) as ws:
            await ws.send(request)
            yield AsyncLiveAgentEngineSession(websocket=ws)
